from kivy.uix.floatlayout import FloatLayout
from kivy.properties import ObjectProperty, StringProperty
from clease.gui.util import parse_select_cond
from ase.db import connect
from copy import deepcopy


class DbBrowser(FloatLayout):
    close = ObjectProperty(None)
    text = StringProperty('')

    def __init__(self, **kwargs):
        self.db = connect(kwargs.pop('db_name'))
        self.spacing = self._format_constraints()
        FloatLayout.__init__(self, **kwargs)
        txt = '| {:{space}s} |'.format('id', space=self.spacing['id'])
        txt += ' {:{space}s} |'.format('formula', space=self.spacing['formula'])
        txt += ' {:{space}s} |'.format('calculator', space=self.spacing['calc'])
        txt += ' {:{space}s} |'.format('energy', space=self.spacing['energy'])
        txt += ' {:{space}s} |'.format('name', space=self.spacing['name'])
        txt += ' {:{space}s} |'.format('size', space=self.spacing['size'])
        txt += ' {:{space}s} |'.format('gen', space=self.spacing['gen'])
        txt += ' {:{space}s} |'.format('struct_type',
                                       space=self.spacing['struct_type'])
        txt += ' {:9s} |'.format('converged')
        self.header = txt
        self.set_rows(None)

    def _format_row(self, row):
        calc = row.get('calculator', '')
        name = row.get('name', 'NN')
        gen = row.get('gen', '')
        energy = row.get('energy', '')
        str_type = row.get('struct_type', '')
        cnv = row.get('converged', '')
        size = row.get('size', '')
        if isinstance(energy, float):
            energy = '{:.3f}'.format(energy)

        if isinstance(cnv, bool):
            cnv = int(cnv)
        txt = '| {:{space}s} |'.format(str(row.id), space=self.spacing['id'])
        txt += ' {:{space}s} |'.format(row.formula,
                                       space=self.spacing['formula'])
        txt += ' {:{space}s} |'.format(calc, space=self.spacing['calc'])
        txt += ' {:{space}s} |'.format(energy, space=self.spacing['energy'])
        txt += ' {:{space}s} |'.format(name, space=self.spacing['name'])
        txt += ' {:{space}s} |'.format(size, space=self.spacing['size'])
        txt += ' {:{space}s} |'.format(str(gen), space=self.spacing['gen'])
        txt += ' {:{space}s} |'.format(str_type,
                                       space=self.spacing['struct_type'])
        txt += ' {:9s} |'.format(str(cnv))
        return txt + '\n'

    def _format_constraints(self):
        spacing = {
            'id': 2,
            'formula': 7,
            'calc': 10,
            'energy': 6,
            'name': 4,
            'gen': 3,
            'struct_type': 11,
            'size': 4
        }

        for row in self.db.select():
            for k, v in spacing.items():
                value = row.get(k, 'none')

                if isinstance(value, float):
                    value = '{:.3f}'.format(value)
                if len(str(value)) > v:
                    spacing[k] = len(str(value))
        return spacing

    def on_select(self, txt):
        select_cond = parse_select_cond(txt)
        self.set_rows(select_cond)

    def set_rows(self, select_cond):
        all_str = deepcopy(self.header) + '\n'

        if select_cond is None:
            db_iterator = self.db.select()
        else:
            db_iterator = self.db.select(select_cond)
        for row in db_iterator:
            str_rep = self._format_row(row)
            all_str += str_rep

        self.text = all_str



from clease import LinearRegression
from clease import Evaluate
from clease import _logger
import numpy as np
import multiprocessing as mp
import os
from random import shuffle, choice
os.environ["OPENBLAS_MAIN_FREE"] = "1"

workers = None


class SaturatedPopulationError(Exception):
    pass


class GAFit(object):
    """
    Genetic Algorithm for selecting relevant clusters.

    Parameters:

    setting: ClusterExpansionSetting
        Setting object used for the cluster expanion

    max_cluster_dia: float
        Maximum diameter included in the population

    max_cluster_size: int
        Maximum number of atoms included in the largest cluster

    elitism: int
        Number of best structures that will be passed unaltered on to the next
        generation

    fname: str
        File name used to backup the population. If this file exists, the next
        run will load the population from the file and start from there.
        Another file named 'fname'_cf_names.txt is created to store the
        names of selected clusters.

    num_individuals: int or str
        Integer with the number of inidivuals or it is equal to "auto",
        in which case 10 times the number of candidate clusters is used

    max_num_in_init_pool: int
        If given the maximum clusters included in the initial population
        is given by this number. If max_num_in_init_pool=150, then
        solution with maximum 150 will be present in the initial pool.

    parallel: bool
        If ``True``, multiprocessing will be used to parallelize over the
        individuals in the population.
        NOTE: One of the most CPU intensive tasks involves matrix
        manipulations using Numpy. If your Numpy installation uses
        hyperthreading, it is possible that running with parallel=True
        actually leads to lower performance.

    num_core: int
        Number of cores to use during parallelization.
        If not given (and parallel=True) then mp.cpu_count()/2
        will be used

    select_cond: list
        Select condition passed to Evaluate to select which
        data points from the database the should be included

    cost_func: str
        Use the inverse as fitness measure.
        Possible cost functions:
        bic - Bayes Information Criterion
        aic - Afaike Information Criterion
        aicc - Modified Afaikes Information Criterion
        (tend to avoid overfitting better than aic)
        loocv - Leave one out cross validation (average)
        max_loocv - Leave one out cross valdition (maximum)
        hqc - Hannan-Quinn information criterion

    sparsity_slope: int
        Ad hoc parameter that can be used to tune the sparsity
        of the model GA selects. The higher it is, the
        sparser models will be prefered. Has only impact
        if cost_func is bic or aic. Default value is 1.

    min_weight: float
        Weight given to the point furthest away from the
        convex hull.

    include_subclusters: bool
        If True all sub-clusters are forced to be part of the solution.
        If a 3-body cluster is included, all the 2-body clusters that
        are part of it will automatically be included.

    Example::

        from clease import Evaluate
        from clease import GAFit

        setting = None # Should be an ASE ClusterExpansionSetting object
        ga_fit = GAFit(setting)
        ga_fit.run()
    """

    def __init__(self, setting=None, max_cluster_size=None,
                 max_cluster_dia=None, mutation_prob=0.001,
                 elitism=1, fname="ga_fit.csv", num_individuals="auto",
                 local_decline=True,
                 max_num_in_init_pool=None, parallel=False, num_core=None,
                 select_cond=None, cost_func="bic", sparsity_slope=1.0,
                 min_weight=1.0, include_subclusters=True):
        evaluator = Evaluate(setting, max_cluster_dia=max_cluster_dia,
                             max_cluster_size=max_cluster_size,
                             select_cond=select_cond, min_weight=min_weight)
        self.setting = setting
        self.eff_num = evaluator.effective_num_data_pts
        self.W = np.diag(evaluator.weight_matrix)
        self.cf_names = evaluator.cf_names
        self.include_subclusters = include_subclusters
        self.sub_constraint = None
        self.super_constraint = None
        if self.include_subclusters:
            self.sub_constraint = self._initialize_sub_cluster_constraint()
            self.super_constraint = self._initialize_super_cluster_constraint()

        allowed_cost_funcs = ["loocv", "bic", "aic", "max_loocv", "aicc",
                              "hqc"]

        if cost_func not in allowed_cost_funcs:
            raise ValueError("Cost func has to be one of {}"
                             "".format(allowed_cost_funcs))

        self.cost_func = cost_func
        self.sparsity_slope = sparsity_slope
        # Read required attributes from evaluate
        self.cf_matrix = evaluator.cf_matrix
        self.cf_names = evaluator.cf_names
        self.e_dft = evaluator.e_dft
        self.fname = fname
        if fname is not None:
            self.fname_cf_names = \
                fname.rpartition(".")[0] + "_cf_names.txt"
        else:
            self.fname_cf_names = None

        if num_individuals == "auto":
            self.pop_size = 10*self.cf_matrix.shape[1]
        else:
            self.pop_size = int(num_individuals)

        # Make sure that the population size is an even number
        if self.pop_size % 2 == 1:
            self.pop_size += 1
        self.num_genes = self.cf_matrix.shape[1]
        self.individuals = self._initialize_individuals(max_num_in_init_pool)
        self.fitness = np.zeros(len(self.individuals))
        self.regression = LinearRegression()
        self.elitism = elitism
        self.mutation_prob = mutation_prob
        self.parallel = parallel
        self.num_core = num_core
        self.statistics = {"best_cv": [],
                           "worst_cv": []}
        self.evaluate_fitness()
        self.local_decline = local_decline
        self.check_valid()

    def _initialize_individuals(self, max_num):
        """Initialize a random population."""
        individuals = []
        if os.path.exists(str(self.fname)):
            individuals = self._init_from_file()
        else:
            max_num = max_num or self.num_genes
            indices = list(range(self.num_genes))
            num_non_zero = np.array(list(range(0, self.pop_size)))
            num_non_zero %= max_num
            num_non_zero[num_non_zero < 3] = 3
            for i in range(self.pop_size):
                shuffle(indices)
                individual = np.zeros(self.num_genes, dtype=np.uint8)
                indx = indices[:num_non_zero[i]]
                individual[np.array(indx)] = 1
                individuals.append(self.make_valid(individual))
        return individuals

    def _init_from_file(self):
        """Initialize the population from file."""
        _logger("Initializing population from {}".format(self.fname))
        individuals = []
        with open(self.fname, 'r') as infile:
            for line in infile:
                individual = np.zeros(self.num_genes, dtype=np.uint8)
                indices = np.array([int(x.strip()) for x in line.split(",")])
                individual[indices] = 1
                individuals.append(individual)
        return individuals

    def bic(self, mse, nsel):
        """Return the Bayes Information Criteria."""
        N = len(self.e_dft)
        sparsity_cost = max((N, self.cf_matrix.shape[1]))
        return N*np.log(mse) + nsel*np.log(sparsity_cost)*self.sparsity_slope

    def aic(self, mse, num_features):
        """Return Afaike information criterion."""
        N = len(self.e_dft)
        return N*np.log(mse) + 2*num_features*self.sparsity_slope

    def hqc(self, mse, num_features):
        """Return Hannan-Quinn information criterion."""
        N = len(self.e_dft)
        return N*np.log(mse) + \
            2*num_features*self.sparsity_slope*np.log(np.log(N))

    def aicc(self, mse, num_features):
        """Modified Afaikes informatiion criterion."""
        aic = self.aic(mse, num_features)
        N = len(self.e_dft)
        corr = 2*num_features**2 + 2*num_features
        corr /= (N - num_features - 1)
        return aic + self.sparsity_slope*corr

    def get_eci(self, individual):
        """Calculate the LOOCV for the current individual."""
        X = self.design_matrix(individual)
        y = self.e_dft
        return self.regression.fit(X, y)

    def design_matrix(self, individual):
        """Return the corresponding design matrix."""
        return self.cf_matrix[:, individual == 1]

    def loo_dev(self, individual):
        """Calculate the prediction error of each data point when left out."""
        coeff = self.get_eci(individual)
        X = self.design_matrix(individual)
        e_pred = X.dot(coeff)
        delta_e = self.e_dft - e_pred
        prec = self.regression.precision_matrix(X)

        denom = 1 - np.diag(X.dot(prec).dot(X.T))
        mask = np.abs(denom) > 1E-7
        denom[~mask] = 1.0
        loo_dev = (self.W*delta_e / denom)**2
        loo_dev[~mask] = 10*np.max(delta_e)
        return loo_dev

    def loocv(self, individual):
        cv_sq = np.sum(self.loo_dev(individual))/self.eff_num

        if cv_sq < 0.0:
            return 0.0
        return 1000.0*np.sqrt(cv_sq)

    def fit_individual(self, individual):
        coeff = self.get_eci(individual)
        X = self.design_matrix(individual)
        e_pred = X.dot(coeff)
        delta_e = self.e_dft - e_pred

        info_measure = None
        n_selected = np.sum(individual)
        mse = np.sum(self.W*delta_e**2)/self.eff_num

        if self.cost_func == "bic":
            info_measure = self.bic(mse, n_selected)
        elif self.cost_func == "aic":
            info_measure = self.aic(mse, n_selected)
        elif self.cost_func == "aicc":
            info_measure = self.aicc(mse, n_selected)
        elif self.cost_func == "hqc":
            info_measure = self.hqc(mse, n_selected)
        elif "loocv" in self.cost_func:
            cv = self.loocv(individual)

            if self.cost_func == "loocv":
                info_measure = cv
            elif self.cost_func == "max_loocv":
                info_measure = np.sqrt(np.max(self.loo_dev(individual)))*1000.0
            else:
                raise ValueError("Unknown LOOCV measure!")
        else:
            raise ValueError("Unknown cost function {}!"
                             "".format(self.cost_func))
        return coeff, info_measure

    def evaluate_fitness(self):
        """Evaluate fitness of all species."""
        global workers

        if self.parallel:
            num_core = self.num_core or int(mp.cpu_count()/2)
            args = [(self, indx) for indx in range(len(self.individuals))]
            if workers is None:
                workers = mp.Pool(num_core)
            self.fitness[:] = workers.map(eval_fitness, args)
        else:
            for i, ind in enumerate(self.individuals):
                _, fit = self.fit_individual(ind)
                self.fitness[i] = -fit

    def flip_one_mutation(self, individual):
        """Apply mutation where one bit flips."""
        indx_sel = list(np.argwhere(individual.T == 1).T[0])
        ns = list(np.argwhere(individual.T == 0).T[0])

        assert len(indx_sel) != 0 or len(ns) != 0
        if len(ns) == 0:
            ns = indx_sel
        elif len(indx_sel) == 0:
            indx_sel = ns

        # Flip included or not included cluster with equal
        # probability
        if np.random.rand() < 0.5:
            indx = choice(indx_sel)
        else:
            indx = choice(ns)
        individual[indx] = (individual[indx] + 1) % 2

        if individual[indx] == 0 and self.include_subclusters:
            name = self.cf_names[indx]
            individual = self._remove_super_clusters(name, individual)
        return individual

    def make_valid(self, individual):
        """Make sure that there is at least two active ECIs."""
        if np.sum(individual) < 2:
            while np.sum(individual) < 2:
                indx = np.random.randint(low=0, high=len(individual))
                individual[indx] = 1

        # Check if subclusters should be included
        if self.include_subclusters:
            individual = self._activate_all_subclusters(individual)
        return individual

    def create_new_generation(self):
        """Create a new generation."""
        new_generation = []
        srt_indx = np.argsort(self.fitness)[::-1]

        assert self.fitness[srt_indx[0]] >= self.fitness[srt_indx[1]]

        # Pass the fittest to the next generation
        num_transfered = 0
        counter = 0
        while num_transfered < self.elitism and counter < len(srt_indx):
            indx = srt_indx[counter]

            individual = self.individuals[indx].copy()

            # Transfer the best
            new_generation.append(individual)

            # Transfer the best individual with a mutation
            new_ind = self.flip_one_mutation(individual.copy())
            new_ind = self.make_valid(new_ind)
            while self._is_in_population(new_ind, new_generation):
                new_ind = self.flip_one_mutation(individual.copy())
                new_ind = self.make_valid(new_ind)

            new_generation.append(new_ind)
            num_transfered += 1
            counter += 1

        if counter >= len(srt_indx):
            self.save_population()
            raise SaturatedPopulationError("The entrie population has "
                                           "saturated!")

        only_positive = self.fitness - np.min(self.fitness)
        cumulative_sum = np.cumsum(only_positive)
        cumulative_sum /= cumulative_sum[-1]
        num_inserted = len(new_generation)

        max_attempts = 100*self.pop_size
        # Create new generation by mergin existing
        for i in range(num_inserted, int(self.pop_size/2)+1):
            rand_num = np.random.rand()
            p1 = np.argmax(cumulative_sum > rand_num)
            p2 = p1
            while p2 == p1:
                rand_num = np.random.rand()
                p2 = np.argmax(cumulative_sum > rand_num)

            new_individual = self.individuals[p1].copy()
            new_individual2 = self.individuals[p2].copy()

            mask = np.random.randint(0, high=2, size=len(new_individual))
            mask = mask.astype(np.uint8)
            new_individual[mask] = self.individuals[p2][mask]
            new_individual2[mask] = self.individuals[p1][mask]
            new_individual = self.make_valid(new_individual)
            new_individual2 = self.make_valid(new_individual2)

            # Check if there are any equal individuals in
            # the population
            counter = 0
            while (self._is_in_population(new_individual, new_generation)
                    and counter < max_attempts):
                new_individual = self.flip_one_mutation(new_individual)
                new_individual = self.make_valid(new_individual)
                counter += 1

            if counter >= max_attempts:
                self.save_population()
                raise SaturatedPopulationError("Popluation is saturated!")
            new_generation.append(new_individual)

            counter = 0
            while (self._is_in_population(new_individual2, new_generation)
                    and counter < max_attempts):
                new_individual2 = self.flip_one_mutation(new_individual2)
                new_individual2 = self.make_valid(new_individual2)
                counter += 1

            if counter >= max_attempts:
                self.save_population()
                raise SaturatedPopulationError("Popluation is saturated!")

            new_generation.append(new_individual2)

        if len(new_generation) != len(self.individuals):
            raise RuntimeError("Size of generation changed! "
                               "Original size: {}. New size: {}"
                               "".format(len(self.individuals),
                                         len(new_generation)))
        self.individuals = new_generation

    def _is_in_population(self, ind, pop):
        """Check if the individual is already in the population."""
        return any(np.all(ind == x) for x in pop)

    def mutate(self):
        """Introduce mutations."""
        avg_f = np.mean(np.abs(self.fitness))
        best_indx = np.argmax(self.fitness)
        for i in range(len(self.individuals)):
            if i == best_indx:
                # Do not mutate the best individual
                continue

            mut_prob = self.mutation_prob

            # Increase the probability of introducing mutations
            # to the least fit individuals
            if abs(self.fitness[i]) > avg_f:
                mut_prob *= abs(self.fitness[i])/avg_f

            if mut_prob > 1.0:
                mut_prob = 1.0

            ind = self.individuals[i].copy()
            mutated = False
            assert mut_prob >= 0.0
            if np.random.rand() < mut_prob:
                ind = self.flip_one_mutation(ind)
                mutated = True

            if mutated:
                self.individuals[i] = self.make_valid(ind)
                _, fit = self.fit_individual(self.individuals[i])
                self.fitness[i] = -fit

    def population_diversity(self):
        """Check the diversity of the population."""
        std = np.std(self.individuals)
        return np.mean(std)

    def log(self, msg):
        """Log messages."""
        _logger(msg)

    @property
    def best_individual(self):
        best_indx = np.argmax(self.fitness)
        individual = self.individuals[best_indx]
        return individual

    @property
    def best_cv(self):
        return 1.0/np.max(self.fitness)

    @property
    def best_individual_indx(self):
        best_indx = np.argmax(self.fitness)
        return best_indx

    @staticmethod
    def get_instance_array():
        raise TypeError("Does not make sense to create an instance array GA.")

    @property
    def selected_cf_names(self):
        from itertools import compress
        individual = self.best_individual
        return list(compress(self.cf_names, individual))

    def index_of_selected_clusters(self, individual):
        """Return the indices of the selected clusters

        Parameters:

        individual: int
            Index of the individual
        """
        return list(np.nonzero(self.individuals[individual])[0])

    def save_population(self):
        # Save population
        self.check_valid()
        if self.fname is None:
            return
        with open(self.fname, 'w') as out:
            for i in range(len(self.individuals)):
                out.write(",".join(str(x) for x in
                                   self.index_of_selected_clusters(i)))
                out.write("\n")
        _logger("\nPopulation written to {}".format(self.fname))

    def save_cf_names(self):
        """Store cluster names of best population to file."""
        if self.fname_cf_names is None:
            return

        with open(self.fname_cf_names, 'w') as out:
            for name in self.selected_cf_names:
                out.write(name+"\n")
        _logger("Selected cluster names saved to "
                "{}".format(self.fname_cf_names))

    def plot_evolution(self):
        """Create a plot of the evolution."""
        from matplotlib import pyplot as plt
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.plot(self.statistics["best_cv"], label="best")
        ax.plot(self.statistics["worst_cv"], label="worst")
        ax.set_xlabel("Generation")
        ax.set_ylabel("CV score (meV/atom)")
        plt.show()

    def run(self, gen_without_change=100, min_change=0.01, save_interval=100):
        """Run the genetic algorithm.

        Return a list consisting of the names of selected clusters at the end
        of the run.

        Parameters:

        gen_without_change: int
            Terminate if gen_without_change are created without sufficient
            improvement

        min_change: float
            Changes a larger than this value is considered "sufficient"
            improvement

        save_interval: int
            Rate at which all the populations are backed up in a file
        """
        num_gen_without_change = 0
        current_best = 0.0
        gen = 0
        while(True):
            self.evaluate_fitness()

            best_indx = np.argmax(self.fitness)

            num_eci = np.sum(self.individuals[best_indx])
            diversity = self.population_diversity()
            self.statistics["best_cv"].append(np.max(self.fitness))
            self.statistics["worst_cv"].append(np.min(self.fitness))

            best3 = np.abs(np.sort(self.fitness)[::-1][:3])
            loocv_msg = ""
            if "loocv" not in self.cost_func:
                # Print the LOOCV
                loocv_msg = "loocv: {:.2f}".format(
                    self.loocv(self.best_individual))

            self.log("Generation: {}. Top 3 {}: {:.2e} (-){:.2e} (-){:.2e} "
                     "Num ECI: {}. Pop. div: {:.2f} {}"
                     "".format(gen, self.cost_func,
                               best3[0], best3[0] - best3[1],
                               best3[0] - best3[2], num_eci, diversity,
                               loocv_msg))
            self.mutate()
            self.create_new_generation()
            if abs(current_best - self.fitness[best_indx]) > min_change:
                num_gen_without_change = 0
            else:
                num_gen_without_change += 1
            current_best = self.fitness[best_indx]

            if gen % save_interval == 0:
                self.save_population()
                self.save_cf_names()

            if num_gen_without_change >= gen_without_change:
                self.log("\nReached {} generations without sufficient "
                         "improvement.".format(gen_without_change))
                break
            gen += 1

        if self.local_decline:
            # Perform a last local optimization
            self._local_optimization()
        self.save_population()
        self.save_cf_names()
        return self.selected_cf_names

    def _local_optimization(self, indx=None):
        """Perform a local optimization strategy to the best individual."""
        from copy import deepcopy
        if indx is None:
            individual = self.best_individual
        else:
            individual = self.individuals[indx]

        num_steps = 100*len(individual)
        self.log("Local optimization with {} trial updates.".format(num_steps))
        cv_min = -np.max(self.fitness)
        self.log("Initial {}: {:.2e}".format(self.cost_func, cv_min))
        for _ in range(num_steps):
            flip_indx = choice(range(len(individual)))
            individual_cpy = deepcopy(individual)
            individual_cpy[flip_indx] = (individual_cpy[flip_indx]+1) % 2
            if individual_cpy[flip_indx] == 0 and self.include_subclusters:
                name = self.cf_names[flip_indx]
                individual_cpy = self._remove_super_clusters(name,
                                                             individual_cpy)
            individual_cpy = self.make_valid(individual_cpy)
            _, cv = self.fit_individual(individual_cpy)

            if cv < cv_min:
                cv_min = cv
                individual = individual_cpy

        for i in range(len(self.individuals)):
            if np.allclose(individual, self.individuals[i]):
                # The individual already exists in the population so we don't
                # insert it
                return

        self.individuals[self.best_individual_indx] = individual
        self.fitness[self.best_individual_indx] = -cv_min
        self.log("Final {}: {:.2e}".format(self.cost_func, cv_min))

    def _initialize_sub_cluster_constraint(self):
        """Initialize the sub-cluster constraint."""
        must_be_active = []
        for name in self.cf_names:
            prefix = name.rpartition("_")[0]
            if prefix == 'c0' or prefix == 'c1' or prefix == '':
                must_be_active.append([])
                continue
            cluster = self.setting.cluster_list.get_by_name(prefix)[0]
            sub = self.setting.cluster_list.get_subclusters(cluster)
            sub_names = [c.name for c in sub]
            indx = []
            for sub_name in sub_names:
                indices = [i for i, name in enumerate(self.cf_names)
                           if name.startswith(sub_name)]
                indx += indices
            must_be_active.append(indx)
        return must_be_active

    def _initialize_super_cluster_constraint(self):
        """Initialize the super-clusters."""
        deactivate = {}
        for name in self.cf_names:
            prefix = name.rpartition("_")[0]
            if prefix in deactivate.keys():
                continue
            indx = []
            for i, name2 in enumerate(self.cf_names):
                prefix2 = name2.rpartition("_")[0]

                if prefix2 == 'c0' or prefix2 == 'c1' or prefix2 == '':
                    continue
                cluster = self.setting.cluster_list.get_by_name(prefix2)[0]
                sub = self.setting.cluster_list.get_subclusters(cluster)
                sub_names = [c.name for c in sub]
                if prefix in sub_names:
                    indx.append(i)
            deactivate[prefix] = indx
        return deactivate

    def _activate_all_subclusters(self, individual):
        """Activate all sub-clusters of the individual."""
        selected_indx = np.nonzero(individual)[0].tolist()
        active = set()
        for indx in selected_indx:
            active = active.union(self.sub_constraint[indx])

        active = list(active)
        if not active:
            return individual
        individual[active] = 1
        return individual

    def _remove_super_clusters(self, name, ind):
        """Remove all the larger clusters."""
        prefix_name = name.rpartition("_")[0]
        ind[self.super_constraint[prefix_name]] = 0
        return ind

    def check_valid(self):
        """Check that the current population is valid."""
        for ind in self.individuals:
            valid = self.make_valid(ind.copy())
            if np.any(valid != ind):
                raise ValueError("Individual violate constraints!")


def eval_fitness(args):
    ga = args[0]
    indx = args[1]
    _, cv = ga.fit_individual(ga.individuals[indx])
    return -cv



# -*- coding: utf-8 -*-
"""Monte Carlo method for ase."""
from __future__ import division
import time
import numpy as np
from ase.units import kB
from clease.montecarlo.exponential_filter import ExponentialFilter
from clease.montecarlo.averager import Averager
from clease.montecarlo import BiasPotential
from clease.montecarlo .swap_move_index_tracker import SwapMoveIndexTracker
from clease import _logger


class DidNotReachEquillibriumError(Exception):
    pass


class TooFewElementsError(Exception):
    pass


class CanNotFindLegalMoveError(Exception):
    pass


class Montecarlo(object):
    """Class for running Monte Carlo at a fixed composition (canonical).

    Parameters:

    atoms: Atoms object
        ASE Atoms object (with CE calculator attached)

    temp: float
        Temperature of Monte Carlo simulation in Kelvin
    """

    def __init__(self, atoms, temp):
        self.name = "MonteCarlo"
        self.atoms = atoms
        self.T = temp

        # List of observers that will be called every n-th step
        # similar to the ones used in the optimization routines
        self.observers = []

        self.constraints = []
        self.max_allowed_constraint_pass_attempts = 1000

        if self.max_allowed_constraint_pass_attempts <= 0:
            raise ValueError("Max. constraint attempts has to be > 0!")

        self.bias_potentials = []

        self.current_step = 0
        self.num_accepted = 0
        self.status_every_sec = 30
        self.atoms_tracker = SwapMoveIndexTracker()
        self.symbols = []
        E0 = self.atoms.get_calculator().calculate(None, None, None)
        self.atoms.get_calculator().clear_history()
        self._build_atoms_list()
        self.current_energy = E0
        self.bias_energy = 0.0
        self.new_bias_energy = self.bias_energy
        self.new_energy = self.current_energy

        # Keep the energy of old and trial state
        self.last_energies = np.zeros(2)
        self.trial_move = []  # Last trial move performed
        self.mean_energy = Averager(ref_value=E0)
        self.energy_squared = Averager(ref_value=E0)
        self.energy_bias = 0.0
        self.update_energy_bias = True

        # Some member variables used to update the atom tracker, only relevant
        # for canonical MC
        self.rand_a = 0
        self.rand_b = 0
        self.selected_a = 0
        self.selected_b = 0
        self.quit = False

        self.filter = ExponentialFilter(min_time=0.2*len(self.atoms),
                                        max_time=20*len(self.atoms),
                                        n_subfilters=10)

    def _probe_energy_bias(self, num_sweeps=2):
        """
        Run MC steps to probe the energy bias. The bias
        will be subtracted off the zeroth ECI and then
        added to the total energy during post processing.
        """
        self.energy_bias = 0.0
        num_steps = num_sweeps*len(self.atoms)
        self.log("Probing energy bias using {} MC steps...".format(num_steps))
        for _ in range(num_steps):
            E, _ = self._mc_step()
            self.mean_energy += E
            self.energy_squared += E**2

        self.log("Energy after probing: {}".format(self.current_energy))
        self.energy_bias = self.current_energy
        self._remove_bias_from_empty_eci(self.energy_bias)

    def _remove_bias_from_empty_eci(self, bias):
        """Remove the energy bias from the empty cluster's ECI.

        Parameters:

        bias: float
            Energy bias
        """
        calc = self.atoms.get_calculator()
        eci = calc.eci
        c0_eci = eci['c0']
        c0_eci -= bias/len(self.atoms)
        eci['c0'] = c0_eci

        calc.update_eci(eci)

        # Force re-calculation of the energy
        self.current_energy = calc.calculate(None, None, None)
        self.last_energies[0] = self.current_energy

        if abs(self.current_energy) > 1E-6:
            raise RuntimeError("Energy is not 0 after subtracting "
                               "the bias. Got {}".format(self.current_energy))

        self.log('Bias subtracted from empty cluster...')

    def _undo_energy_bias_from_eci(self):
        eci = self.atoms.get_calculator().eci
        eci['c0'] += self.energy_bias/len(self.atoms)
        self.atoms.get_calculator().update_eci(eci)
        calc = self.atoms.get_calculator()
        self.current_energy = calc.calculate(None, None, None)
        #self.energy_bias = 0.0
        self.log('Empty cluster ECI reset to original value...')

    def insert_symbol(self, symb, indices):
        """Insert symbols on a predefined set of indices.

        Parameters:

        symb: str
            Symbol to be inserted
        indices: list
            Indices where symb should be inserted
        """
        calc = self.atoms.get_calculator()
        for indx in indices:
            system_changes = (indx, self.atoms[indx].symbol, symb)

            if not self._no_constraint_violations([system_changes]):
                raise ValueError("The indices given results in an update "
                                 "that violate one or more constraints!")
            calc.update_cf(system_changes)
        self._build_atoms_list()
        calc.clear_history()

    def insert_symbol_random_places(self, symbol, num=1, swap_symbs=[]):
        """Insert random symbol.

        Parameters:

        symbol: str
            Symbol to insert

        num: int
            Number of sites to insert

        swap_symbs: list
            If given, will insert the replace symbol with sites having symbols
            in this list
        """
        from random import choice
        if not swap_symbs:
            symbs = [atom.symbol for atom in self.atoms]
            swap_symbs = list(set(symbs))
        num_inserted = 0
        max_attempts = 10 * len(self.atoms)

        calc = self.atoms.get_calculator()
        attempts = 0
        while num_inserted < num and attempts < max_attempts:
            attempts += 1
            old_symb = choice(swap_symbs)
            if old_symb == symbol:
                continue

            indx = self.atoms_tracker.get_random_indx_of_symbol(old_symb)
            if self.atoms[indx].symbol not in swap_symbs:
                # This can happen because the atom_indx is inconsistent
                # until after all the atoms have been inserted
                continue

            system_changes = (indx, old_symb, symbol)
            if not self._no_constraint_violations([system_changes]):
                continue
            calc.update_cf(system_changes)
            num_inserted += 1
        self._build_atoms_list()
        calc.clear_history()

        if attempts == max_attempts:
            raise RuntimeError("Could insert {} {} atoms!".format(num, symbol))

    def update_current_energy(self):
        """Enforce a new energy evaluation."""
        self.current_energy = \
            self.atoms.get_calculator().calculate(None, None, None)
        self.bias_energy = 0.0
        for bias in self.bias_potentials:
            self.bias_energy += bias.calculate_from_scratch(self.atoms)
        self.current_energy += self.bias_energy

    def set_symbols(self, symbs):
        """Set the symbols of this Monte Carlo run.

        Parameters:

        symbs: list
            Symbols to insert. Has to have the same length as the
            attached atoms object.
        """
        self.atoms.get_calculator().set_symbols(symbs)
        self._build_atoms_list()
        self.update_current_energy()

    def _check_symbols(self):
        """Check that there is at least two different symbols."""
        symbs = [atom.symbol for atom in self.atoms]
        count = {}
        for symb in symbs:
            if symb not in count:
                count[symb] = 1
            else:
                count[symb] += 1

        # Verify that there is at two elements with more that two symbols
        if len(count.keys()) < 2:
            raise TooFewElementsError(
                "There is only one element in the given atoms object!")
        n_elems_more_than_2 = 0
        for _, value in count.items():
            if value >= 2:
                n_elems_more_than_2 += 1
        if n_elems_more_than_2 < 2:
            raise TooFewElementsError(
                "There is only one element that has more than one atom")

    def log(self, msg, mode="info"):
        """Logs the message as info."""
        allowed_modes = ["info", "warning"]
        if mode not in allowed_modes:
            raise ValueError("Mode has to be one of {}".format(allowed_modes))

        _logger(msg)

    def _no_constraint_violations(self, system_changes):
        """Check if the proposed moves violate any of the constraints.

        Parameters:

        system_changes: list
            Changes of the proposed move
        """
        for constraint in self.constraints:
            if not constraint(system_changes):
                return False
        return True

    def reset(self):
        """Reset all member variables to their original values."""
        for _, obs in self.observers:
            obs.reset()

        self.filter.reset()
        self.current_step = 0
        self.num_accepted = 0
        self.mean_energy.clear()
        self.energy_squared.clear()
        self.corrtime_energies = []

    def _build_atoms_list(self):
        """
        Create a dictionary of the indices of each atom which is used to
        make sure that two equal atoms cannot be swapped
        """
        self.atoms_tracker.init_tracker(self.atoms)
        self.symbols = self.atoms_tracker.symbols

    def _update_tracker(self, system_changes):
        """Update the atom tracker."""
        self.atoms_tracker.update_swap_move(system_changes)

    def add_constraint(self, constraint):
        """Add a new constraint to the sampler.

        Parameters:

        constraint: MCConstraint
            Instance of a constraint object
        """
        self.constraints.append(constraint)

    def add_bias(self, potential):
        """Add a new bias potential.

        Parameters:

        potential: Bias potential
            Potential to be added
        """
        if not isinstance(potential, BiasPotential):
            raise TypeError("potential has to be of type BiasPotential")
        self.bias_potentials.append(potential)

    def attach(self, obs, interval=1):
        """Attach observers to be called on a given MC step interval.

        Parameters:

        obs: MCObserver
            Observer to be added

        interval: int
            How often the observer should be called
        """
        if not obs.interval_ok(interval):
            name = type(obs).__name__
            raise ValueError("Invalid interval for {}. Check docstring "
                             "of the observer.".format(name))

        self.observers.append((interval, obs))

    def run(self, steps=100):
        """Run Monte Carlo simulation.

        Parameters:

        steps: int
            Number of steps in the MC simulation
        """
        # Check the number of different elements are correct to avoid
        # infinite loops
        self._check_symbols()

        self.update_current_energy()

        # Atoms object should have attached calculator
        # Add check that this is show
        self._mc_step()

        # self.current_step gets updated in the _mc_step function
        self.reset()

        # Probe bias energy and remove bias
        self._probe_energy_bias()
        self.reset()

        start = time.time()
        prev = self.current_step
        while(self.current_step < steps):
            E, _ = self._mc_step()

            self.mean_energy += E
            self.energy_squared += E**2

            if (time.time() - start > self.status_every_sec):
                ms_per_step = 1000.0 * self.status_every_sec / \
                    float(self.current_step - prev)
                accept_rate = self.num_accepted / float(self.current_step)
                self.log(
                    "%d of %d steps. %.2f ms per step. Acceptance rate: %.2f" %
                    (self.current_step, steps, ms_per_step, accept_rate))
                prev = self.current_step
                start = time.time()

            if self.quit:
                break

        self.log(
            "Reached maximum number of steps ({} mc steps)".format(steps))

        # NOTE: Also update current_energy
        self._undo_energy_bias_from_eci()

    @property
    def meta_info(self):
        """Return dict with meta info."""
        import sys
        import datetime
        ts = time.time()
        st = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S')
        v_info = sys.version_info
        meta_info = {
            "timestamp": st,
            "python_version": "{}.{}.{}".format(v_info.major, v_info.minor,
                                                v_info.micro)
        }
        return meta_info

    def get_thermodynamic_quantities(self):
        """Compute thermodynamic quantities."""
        quantities = {}
        mean_energy = self.mean_energy.mean
        quantities["energy"] = mean_energy + self.energy_bias
        mean_sq = self.energy_squared.mean
        quantities["heat_capacity"] = \
            (mean_sq - mean_energy**2) / (kB * self.T**2)
        quantities["energy_var"] = mean_sq - mean_energy**2
        quantities["temperature"] = self.T
        at_count = self.count_atoms()
        for key, value in at_count.items():
            name = "{}_conc".format(key)
            conc = float(value) / len(self.atoms)
            quantities[name] = conc

        # Add some more info that can be useful
        quantities.update(self.meta_info)

        # Add information from observers
        for obs in self.observers:
            quantities.update(obs[1].get_averages())
        return quantities

    def _get_trial_move(self):
        """
        Perform a trial move by swapping two atoms

        """
        n = len(self.atoms)
        self.rand_a = np.random.randint(0, n)
        self.rand_b = np.random.randint(0, n)
        symb_a = self.symbols[np.random.randint(0, len(self.symbols))]
        symb_b = symb_a
        while (symb_b == symb_a):
            symb_b = self.symbols[np.random.randint(0, len(self.symbols))]

        rand_pos_a = self.atoms_tracker.get_random_indx_of_symbol(symb_a)
        rand_pos_b = self.atoms_tracker.get_random_indx_of_symbol(symb_b)
        system_changes = [(rand_pos_a, symb_a, symb_b),
                          (rand_pos_b, symb_b, symb_a)]
        return system_changes

    def _accept(self, system_changes):
        """Return `True` if the trial step is accepted.

        Parameters:

        system_changes: list
            List with system changes
        """
        self.last_energies[0] = self.current_energy

        # NOTE: Calculate updates the system
        calc = self.atoms.get_calculator()
        self.new_energy = calc.get_energy_given_change(system_changes)

        # NOTE: As this is called after calculate, the changes has
        # already been introduced to the system
        self.new_bias_energy = 0.0
        for bias in self.bias_potentials:
            self.new_bias_energy += bias(system_changes)
        self.new_energy += self.new_bias_energy
        self.last_energies[1] = self.new_energy

        # Standard Metropolis acceptance criteria
        if (self.new_energy < self.current_energy):
            return True
        kT = self.T * kB
        energy_diff = self.new_energy - self.current_energy
        probability = np.exp(-energy_diff / kT)
        return np.random.rand() <= probability

    def count_atoms(self):
        """Count the number of each element."""
        atom_count = {key: 0 for key in self.symbols}
        for atom in self.atoms:
            atom_count[atom.symbol] += 1
        return atom_count

    def _mc_step(self):
        """Make one Monte Carlo step by swithing two atoms."""
        self.current_step += 1
        self.last_energies[0] = self.current_energy

        system_changes = self._get_trial_move()
        counter = 0
        while not self._no_constraint_violations(system_changes) and \
                counter < self.max_allowed_constraint_pass_attempts:
            system_changes = self._get_trial_move()
            counter += 1

        self.trial_move = system_changes

        if counter == self.max_allowed_constraint_pass_attempts:
            msg = "Did not manage to produce a trial move that does not "
            msg += "violate any of the constraints"
            raise CanNotFindLegalMoveError(msg)

        move_accepted = self._accept(system_changes)

        # At this point the new energy is calculated in the _accept function
        self.last_energies[1] = self.new_energy

        if (move_accepted):
            self.current_energy = self.new_energy
            self.bias_energy = self.new_bias_energy
            self.num_accepted += 1
        else:
            # Reset the sytem back to original
            for change in system_changes:
                indx = change[0]
                old_symb = change[1]
                assert (self.atoms[indx].symbol == change[2])
                self.atoms[indx].symbol = old_symb

        if (move_accepted):
            self.atoms.get_calculator().clear_history()
        else:
            self.atoms.get_calculator().restore()

        if (move_accepted):
            # Update the atom_indices
            self._update_tracker(system_changes)
        else:
            new_symb_changes = []
            for change in system_changes:
                new_symb_changes.append((change[0], change[1], change[1]))
            system_changes = new_symb_changes

        # Execute all observers
        for entry in self.observers:
            interval = entry[0]
            if (self.current_step % interval == 0):
                obs = entry[1]
                obs(system_changes)
        self.filter.add(self.current_energy)
        return self.current_energy, move_accepted



from enum import IntEnum
from ase.utils import convert_string_to_fd


class LogVerbosity(IntEnum):
    # IntEnum for comparisons
    ERROR = 0
    WARNING = 1
    INFO = 2
    DEBUG = 3


class CLEASELogger:
    """Class for handling printing in CLEASE"""

    def __init__(self, fd='-', verbosity=LogVerbosity.INFO):
        self.oldfd = None           # Dummy
        self.fd = fd                # File descriptor
        self.verbosity = verbosity  # Print messages with verbose <= verbosity

    @property
    def verbosity(self):
        return self._verbosity

    @verbosity.setter
    def verbosity(self, verbosity):
        self._verbosity = self.format_verbose(verbosity)

    @property
    def fd(self):
        return self._fd

    @fd.setter
    def fd(self, fd):
        if fd == self.oldfd:
            return
        self.oldfd = fd
        self._fd = convert_string_to_fd(fd)

    def format_verbose(self, verbose):
        """Ensure that verbose is LogVerbosity type"""
        if isinstance(verbose, LogVerbosity):
            return verbose
        elif isinstance(verbose, int):
            return LogVerbosity(verbose)
        else:
            raise ValueError('Bad verbosity level: {}'.format(verbose))

    def __call__(self, *args, verbose=LogVerbosity.INFO, **kwargs):
        verbose = self.format_verbose(verbose)
        if verbose <= self.verbosity:
            print(*args, file=self._fd, **kwargs)

    def __str__(self):
        return 'Verbosity threshold: {}, FD: {}'.format(str(self.verbosity),
                                                        self.fd)



from clease.cleaselogger import CLEASELogger, LogVerbosity
_logger = CLEASELogger()


def set_verbosity(verbosity):
    _logger.verbosity = verbosity


def set_fd(fd):
    _logger.fd = fd

__version__ = '0.9.11'

from clease.template_filters import SkewnessFilter, EquivalentCellsFilter
from clease.template_filters import ValidConcentrationFilter
from clease.template_filters import DistanceBetweenFacetsFilter
from clease.template_filters import VolumeToSurfaceRatioFilter
from clease.template_filters import AngleFilter
from clease.atoms_manager import AtomsManager
from clease.settings import ClusterExpansionSetting
from clease.concentration import Concentration
from clease.settings_bulk import CEBulk, CECrystal, settingFromJSON
from clease.evaluate import Evaluate
from clease.corrFunc import CorrFunction
from clease.newStruct import NewStructures
from clease.convexhull import ConvexHull
from clease.regression import LinearRegression, Tikhonov, Lasso
from clease.ga_fit import GAFit
from clease.bayesian_compressive_sensing import BayesianCompressiveSensing
from clease.sequential_cluster_ridge import SequentialClusterRidge
from clease.cluster_info_mapper import ClusterInfoMapper
from clease.cluster_manager import ClusterManager

__all__ = ['CEBulk', 'CECrystal', 'Concentration', 'CorrFunction',
           'NewStructures', 'NewStructures', 'Evaluate', 'AtomsManager',
           'ConvexHull', 'LinearRegression', 'Tikhonov', 'Lasso',
           'GAFit', 'BayesianCompressiveSensing', 'LogVerbosity',
           'SkewnessFilter']



from kivy.uix.floatlayout import FloatLayout
from kivy.properties import ObjectProperty
from clease import __version__ as version
import os
import json


class ConcObsEditor(FloatLayout):
    close = ObjectProperty(None)
    backup_folder = '.cleaseGUI/'
    backup_file = 'concObsBackup{}.json'.format(version)
    data = {}
    default_element = 'X'

    def __init__(self, **kwargs):
        self.default_element = kwargs.pop('default_element')
        FloatLayout.__init__(self, **kwargs)
        if not os.path.exists(self.backup_folder):
            os.mkdir(self.backup_folder)
        self.load_values()
        self._init_fields()

    @property
    def fname(self):
        return self.backup_folder + self.backup_file

    def load_values(self):
        if os.path.exists(self.fname):
            with open(self.fname, 'r') as f:
                self.data = json.load(f)

    def _init_fields(self):
        element = self.data.get('element', self.default_element)
        if element == '':
            element = self.default_element

        self.ids.elementInput.text = element

    def backup(self):
        self.data['element'] = self.ids.elementInput.text
        with open(self.fname, 'w') as f:
            json.dump(self.data, f)



class MCObserver(object):
    """Base class for all MC observers."""

    def __init__(self):
        self.name = "GenericObserver"

    def __call__(self, system_changes):
        """
        Gets information about the system changes and can perform some action

        Parameters:

        system_changes: list
            List of system changes. For example, if the occupation of the
            atomic index 23 is changed from Mg to Al,
            system_change = [(23, Mg, Al)].
            If an Mg atom occupying the atomic index 26 is swapped with an Al
            atom occupying the atomic index 12,
            system_change = [(26, Mg, Al), (12, Al, Mg)]
        """
        pass

    def reset(self):
        """Reset all values of the MC observer"""
        pass

    def get_averages(self):
        """Return averages in the form of a dictionary."""
        return {}

    def calculate_from_scratch(self, atoms):
        """
        Method for calculating the tracked value from scratch
        (i.e. without using fast update methods)
        """
        pass

    def interval_ok(self, interval):
        """
        Check if the interval specified on attach is ok. Default is that all
        intervals are OK

        Parameter:

        interval: int
            Interval controlling how often a MC observer will be called.
        """
        return True



"""Logger that can be used together with multiprocessing funcions."""
import logging as lg
import threading
import multiprocessing as mp


class MultiprocessHandler(lg.Handler):
    """Logger class for multiprocessing functions.

    Inspired by Matt Gathu's blog post
    https://mattgathu.github.io/multiprocessing-logging-in-python/
    """

    def __init__(self, fname):
        """Multiprocessing logger."""
        lg.Handler.__init__(self)

        self._handler = lg.FileHandler(fname)
        self.queue = mp.Queue(-1)

        thread = threading.Thread(target=self.receive)
        thread.daemon = True
        thread.start()

    def setFormatter(self, fmt):
        """Set logging format."""
        lg.Handler.setFormatter(self, fmt)
        self._handler.setFormatter(fmt)

    def receive(self):
        """Receive message from queue."""
        while True:
            try:
                record = self.queue.get()
                self._handler.emit(record)
            except (KeyboardInterrupt, SystemExit):
                raise
            except EOFError:
                break
            except Exception as exc:
                print("An unexpected exception occured during logging.")
                print(str(exc))

    def send(self, msg):
        """Send message."""
        self.queue.put_nowait(msg)

    def emit(self, record):
        """Emit record."""
        try:
            self.send(record)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            self.handleError(record)

    def close(self):
        """Close file handler."""
        self._handler.close()
        lg.Handler.close(self)
